﻿' ----------------------------------------------------------------------------------
' Microsoft Developer & Platform Evangelism
' 
' Copyright (c) Microsoft Corporation. All rights reserved.
' 
' THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
' EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
' OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
' ----------------------------------------------------------------------------------
' The example companies, organizations, products, domain names,
' e-mail addresses, logos, people, places, and events depicted
' herein are fictitious.  No association with any real company,
' organization, product, domain name, email address, logo, person,
' places, or events is intended or should be inferred.
' ----------------------------------------------------------------------------------


Imports Microsoft.VisualBasic

Imports System.Globalization
Imports System.Linq
Imports System.Security.Principal
Imports System.Web.Mvc
Imports System.Web.Security
Imports MyTodo.WebUx.Models

Namespace Controllers

    <HandleError()> _
    Public Class AccountController
        Inherits Controller
        ' This constructor is used by the MVC framework to instantiate the controller using
        ' the default forms authentication and membership providers.
        Public Sub New()
            Me.New(Nothing, Nothing, Nothing)
        End Sub

        ' This constructor is not used by the MVC framework but is instead provided for ease
        ' of unit testing this type. See the comments at the end of this file for more
        ' information.
        Public Sub New(ByVal formsAuth As IFormsAuthentication, ByVal membershipService As IMembershipService, ByVal taskRepository As TaskRepository)
            Me.FormsAuth = If(formsAuth, New FormsAuthenticationService())
            Me.MembershipService = If(membershipService, New AccountMembershipService())
            Me.TaskRepository = If(taskRepository, New TaskRepository())
        End Sub

        Private privateFormsAuth As IFormsAuthentication
        Public Property FormsAuth() As IFormsAuthentication
            Get
                Return privateFormsAuth
            End Get
            Private Set(ByVal value As IFormsAuthentication)
                privateFormsAuth = value
            End Set
        End Property

        Private privateMembershipService As IMembershipService
        Public Property MembershipService() As IMembershipService
            Get
                Return privateMembershipService
            End Get
            Private Set(ByVal value As IMembershipService)
                privateMembershipService = value
            End Set
        End Property

        Private privateTaskRepository As TaskRepository
        Public Property TaskRepository() As TaskRepository
            Get
                Return privateTaskRepository
            End Get
            Private Set(ByVal value As TaskRepository)
                privateTaskRepository = value
            End Set
        End Property

        Public Function LogOn() As ActionResult
            Return View()
        End Function

        <AcceptVerbs(HttpVerbs.Post), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1054:UriParametersShouldNotBeStrings", Justification:="Needs to take same parameter type as Controller.Redirect()")> _
        Public Function LogOn(ByVal userName As String, ByVal password As String, ByVal rememberMe As Boolean, ByVal returnUrl As String) As ActionResult
            If (Not Me.ValidateLogOn(userName, password)) Then
                Return View()
            End If

            Me.FormsAuth.SignIn(userName, rememberMe)
            If (Not String.IsNullOrEmpty(returnUrl)) Then
                Return Redirect(returnUrl)
            Else
                Return RedirectToAction("Index", "Task")
            End If
        End Function

        Public Function LogOff() As ActionResult
            Me.FormsAuth.SignOut()

            Return RedirectToAction("Index", "Task")
        End Function

        Public Function Register() As ActionResult
            Me.ViewData("PasswordLength") = Me.MembershipService.MinPasswordLength

            Return View()
        End Function

        <AcceptVerbs(HttpVerbs.Post)> _
        Public Function Register(ByVal userName As String, ByVal email As String, ByVal password As String, ByVal confirmPassword As String) As ActionResult
            Me.ViewData("PasswordLength") = Me.MembershipService.MinPasswordLength

            ' we are only letting a single user register on this site (ever)
            If (Not Roles.RoleExists("Owner")) Then
                If Me.ValidateRegistration(userName, email, password, confirmPassword) Then
                    ' Attempt to register the user
                    Dim createStatus As MembershipCreateStatus = Me.MembershipService.CreateUser(userName, password, email)

                    If createStatus = MembershipCreateStatus.Success Then
                        ' Provision the user role and tables.
                        Roles.CreateRole("Owner")
                        Roles.AddUserToRole(userName, "Owner")

                        Me.TaskRepository.CreateTables()

                        Me.FormsAuth.SignIn(userName, False)

                        ' Provision the Public Lists
                        Me.CreateInitialLists()

                        Return RedirectToAction("Index", "Task")
                    Else
                        ModelState.AddModelError("_FORM", ErrorCodeToString(createStatus))
                    End If
                End If
            Else
                ModelState.AddModelError("_FORM", "The Owner has already registered")
            End If

            ' If we got this far, something failed, redisplay form
            Return View()
        End Function

        <Authorize()> _
        Public Function Profile() As ActionResult
            Me.ViewData("PasswordLength") = Me.MembershipService.MinPasswordLength

            Return View()
        End Function

        <AcceptVerbs(HttpVerbs.Post), Authorize(), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification:="Exceptions result in password not being changed.")> _
        Public Function Profile(ByVal currentPassword As String, ByVal newPassword As String, ByVal confirmPassword As String) As ActionResult
            Me.ViewData("PasswordLength") = Me.MembershipService.MinPasswordLength

            If (Not Me.ValidateChangePassword(currentPassword, newPassword, confirmPassword)) Then
                Return View()
            End If

            Try
                If (Not String.IsNullOrEmpty(newPassword)) Then
                    If (Not Me.MembershipService.ChangePassword(User.Identity.Name, currentPassword, newPassword)) Then
                        ModelState.AddModelError("_FORM", "The current password is incorrect or the new password is invalid.")
                        Return Me.View()
                    End If
                End If

                Return Me.View("ProfileUpdateSuccess")
            Catch
                ModelState.AddModelError("_FORM", "The current password is incorrect or the new password is invalid.")
                Return View()
            End Try
        End Function

        Public Function ProfileUpdateSuccess() As ActionResult
            Return View()
        End Function

        Protected Overrides Sub OnActionExecuting(ByVal filterContext As ActionExecutingContext)
            If TypeOf filterContext.HttpContext.User.Identity Is WindowsIdentity Then
                Throw New InvalidOperationException("Windows authentication is not supported.")
            End If
        End Sub

        Protected Sub CreateInitialLists()
            Dim document = Me.TaskRepository.RetrieveInitialLists()

            If document.Nodes().Count() = 0 Then
                Return
            End If

            Dim todoController = New TodoController()
            todoController.ControllerContext = Me.ControllerContext

            For Each taskList In document.Descendants("TaskList")
                Dim taskListName = taskList.Element("Name").Value
                Dim taskListIsPublic = Boolean.Parse(taskList.Element("IsPublic").Value)

                Dim createdTaskList = Me.TaskRepository.CreateList(taskListName, taskListIsPublic)

                If createdTaskList Is Nothing OrElse String.IsNullOrEmpty(createdTaskList.ListId) Then
                    Return
                End If

                For Each task In taskList.Descendants("Task")
                    Dim newTask = New Task With {.ListId = createdTaskList.ListId, .Subject = task.Element("Subject").Value, .DueDate = If(String.IsNullOrEmpty(task.Element("DueDate").Value), DateTime.MaxValue, DateTime.Parse(task.Element("DueDate").Value, CultureInfo.InvariantCulture)), .IsComplete = Boolean.Parse(task.Element("IsComplete").Value)}

                    todoController.CreateTask(newTask)
                Next task
            Next taskList
        End Sub

#Region "Validation Methods"

        Private Shared Function ErrorCodeToString(ByVal createStatus As MembershipCreateStatus) As String
            ' See http://msdn.microsoft.com/en-us/library/system.web.security.membershipcreatestatus.aspx for
            ' a full list of status codes.
            Select Case createStatus
                Case MembershipCreateStatus.DuplicateUserName
                    Return "Username already exists. Please enter a different user name."

                Case MembershipCreateStatus.DuplicateEmail
                    Return "A username for that e-mail address already exists. Please enter a different e-mail address."

                Case MembershipCreateStatus.InvalidPassword
                    Return "The password provided is invalid. Please enter a valid password value."

                Case MembershipCreateStatus.InvalidEmail
                    Return "The e-mail address provided is invalid. Please check the value and try again."

                Case MembershipCreateStatus.InvalidAnswer
                    Return "The password retrieval answer provided is invalid. Please check the value and try again."

                Case MembershipCreateStatus.InvalidQuestion
                    Return "The password retrieval question provided is invalid. Please check the value and try again."

                Case MembershipCreateStatus.InvalidUserName
                    Return "The user name provided is invalid. Please check the value and try again."

                Case MembershipCreateStatus.ProviderError
                    Return "The authentication provider returned an error. Please verify your entry and try again. If the problem persists, please contact your system administrator."

                Case MembershipCreateStatus.UserRejected
                    Return "The user creation request has been canceled. Please verify your entry and try again. If the problem persists, please contact your system administrator."

                Case Else
                    Return "An unknown error occurred. Please verify your entry and try again. If the problem persists, please contact your system administrator."
            End Select
        End Function

        Private Function ValidateChangePassword(ByVal currentPassword As String, ByVal newPassword As String, ByVal confirmPassword As String) As Boolean
            If (Not String.IsNullOrEmpty(currentPassword)) Then
                If newPassword Is Nothing OrElse newPassword.Length < Me.MembershipService.MinPasswordLength Then
                    ModelState.AddModelError("newPassword", String.Format(CultureInfo.CurrentCulture, "You must specify a new password of {0} or more characters.", Me.MembershipService.MinPasswordLength))
                End If

                If (Not String.Equals(newPassword, confirmPassword, StringComparison.Ordinal)) Then
                    ModelState.AddModelError("_FORM", "The new password and confirmation password do not match.")
                End If
            End If

            Return ModelState.IsValid
        End Function

        Private Function ValidateLogOn(ByVal userName As String, ByVal password As String) As Boolean
            If String.IsNullOrEmpty(userName) Then
                ModelState.AddModelError("username", "You must specify a username.")
            End If

            If String.IsNullOrEmpty(password) Then
                ModelState.AddModelError("password", "You must specify a password.")
            End If

            If (Not Me.MembershipService.ValidateUser(userName, password)) Then
                ModelState.AddModelError("_FORM", "The username or password provided is incorrect.")
            End If

            Return ModelState.IsValid
        End Function

        Private Function ValidateRegistration(ByVal userName As String, ByVal email As String, ByVal password As String, ByVal confirmPassword As String) As Boolean
            If String.IsNullOrEmpty(userName) Then
                ModelState.AddModelError("username", "You must specify a username.")
            End If

            If String.IsNullOrEmpty(email) Then
                ModelState.AddModelError("email", "You must specify an email address.")
            End If

            If password Is Nothing OrElse password.Length < Me.MembershipService.MinPasswordLength Then
                ModelState.AddModelError("password", String.Format(CultureInfo.CurrentCulture, "You must specify a password of {0} or more characters.", Me.MembershipService.MinPasswordLength))
            End If

            If (Not String.Equals(password, confirmPassword, StringComparison.Ordinal)) Then
                ModelState.AddModelError("_FORM", "The new password and confirmation password do not match.")
            End If

            Return ModelState.IsValid
        End Function

#End Region
    End Class
End Namespace