﻿' ----------------------------------------------------------------------------------
' Microsoft Developer & Platform Evangelism
' 
' Copyright (c) Microsoft Corporation. All rights reserved.
' 
' THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
' EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
' OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
' ----------------------------------------------------------------------------------
' The example companies, organizations, products, domain names,
' e-mail addresses, logos, people, places, and events depicted
' herein are fictitious.  No association with any real company,
' organization, product, domain name, email address, logo, person,
' places, or events is intended or should be inferred.
' ----------------------------------------------------------------------------------


Imports MyTodo.WebUx.Helper
Imports Microsoft.VisualBasic

Imports System
Imports System.Collections.Generic
Imports System.IO
Imports System.Linq
Imports System.Net
Imports System.Text
Imports System.Web.Mvc
Imports System.Web.Routing
Imports MyTodo.WebUx.Models

Namespace Controllers

    <CacheFilter(Duration:=-1)> _
    Public Class TodoController
        Inherits Controller
        Private ReadOnly repository As TaskRepository

        Public Sub New()
            Me.New(Nothing)
        End Sub

        Public Sub New(ByVal repository As TaskRepository)
            Me.repository = If(repository, New TaskRepository())
        End Sub

        Public Function GetTasks(ByVal id As String) As ActionResult
            Dim list As TaskList = Me.repository.GetTaskList(id)

            If list IsNot Nothing Then
                Dim tasks As PagedTaskSource = Me.repository.GetTasks(id, 1000, Nothing, Nothing)
                Return Json(tasks, JsonRequestBehavior.AllowGet)
            End If

            Return Nothing
        End Function

        <AcceptVerbs(HttpVerbs.Post), Authorize(Roles:="Owner")> _
        Public Function CreateTask(ByVal task As Task) As ActionResult
            If task.StartDate = DateTime.MaxValue Then
                task.StartDate = DateTime.Now
            End If

            Me.repository.CreateTask(task)

            Return Json(task)
        End Function

        <AcceptVerbs(HttpVerbs.Put), Authorize(Roles:="Owner")> _
        Public Function UpdateTask(ByVal task As Task) As ActionResult
            Dim list As TaskList = Me.repository.GetTaskList(task.ListId)
            Dim oldTask As Task = Me.repository.GetTask(task.TaskId, list.ListId)

            Me.repository.UpdateTask(task)

            Return Json(task)
        End Function

        <AcceptVerbs(HttpVerbs.Delete), Authorize(Roles:="Owner")> _
        Public Function DeleteTask(ByVal task As Task) As ActionResult
            Me.repository.DeleteTask(task.ListId, task.TaskId)

            Dim list As TaskList = Me.repository.GetTaskList(task.ListId)

            Return Json(True)
        End Function

        Public Function GetLists() As ActionResult
            Dim lists As IEnumerable(Of TaskList) = New List(Of TaskList)()

            If Me.repository.TablesExist() Then
                Dim userIsInRole As Boolean = User.IsInRole("Owner")

                lists = Me.repository.GetTaskLists(userIsInRole)
            End If

            Return Json(lists, JsonRequestBehavior.AllowGet)
        End Function

        <AcceptVerbs(HttpVerbs.Post), Authorize(Roles:="Owner")> _
        Public Function CreateList(ByVal taskList As TaskList) As ActionResult
            Dim list As TaskList = Me.repository.CreateList(taskList.Name, taskList.IsPublic)

            Return Json(list)
        End Function

        <AcceptVerbs(HttpVerbs.Put), Authorize(Roles:="Owner")> _
        Public Function UpdateList(ByVal list As TaskList) As ActionResult
            Dim taskList As TaskList = Me.repository.GetTaskList(list.ListId)

            Me.repository.UpdateList(list)

            Return Json(list)
        End Function

        <AcceptVerbs(HttpVerbs.Delete), Authorize(Roles:="Owner")> _
        Public Function DeleteList(ByVal taskList As TaskList) As ActionResult
            Me.repository.DeleteList(taskList.ListId)

            Return Json(True)
        End Function

        Public Function Rss(ByVal listId As String) As ActionResult
            If String.IsNullOrEmpty(listId) Then
                Dim lists As IEnumerable(Of TaskList) = Me.repository.GetTaskLists(User.IsInRole("Owner"))

                Dim feedItems As IList(Of FeedItem) = lists.Select(Function(p) New FeedItem() With {.Title = p.Name, .Description = If(p.IsPublic, "Public list.", "Private list"), .Url = New Uri(Me.GetAbsoluteUrl(p.ListId))}).ToList()

                Dim feed = New Feed(feedItems) With {.Title = "Lists"}

                Return Me.View(feed)
            Else
                Dim list As TaskList = Me.repository.GetTaskList(listId)

                If list IsNot Nothing AndAlso (Request.IsAuthenticated OrElse list.IsPublic) Then
                    Dim tasks As PagedTaskSource = Me.repository.GetTasks(listId, 1000, Nothing, Nothing)

                    Dim feedItems As IList(Of FeedItem) = tasks.Tasks.Select(Function(p) New FeedItem With {.Title = p.Subject, .Description = String.Concat(If(p.IsComplete, "The task is completed", "The task is pending"), If(((Not p.IsComplete) AndAlso p.DueDate < DateTime.Now), " and overdue.", ".")), .Url = New Uri(Me.GetAbsoluteUrl(listId)), .Published = p.StartDate}).ToList()

                    Dim feed As New Feed(feedItems) With {.Title = list.Name, .Description = "Is Public: " & list.IsPublic, .Url = New Uri(Me.GetAbsoluteUrl(listId))}

                    Return Me.View(feed)
                End If

                Return New EmptyResult()
            End If
        End Function

        Protected Overridable Function GetAbsoluteUrl(ByVal listId As String) As String
            Dim path As String = LinkBuilder.BuildUrlFromExpression(Of HomeController)(Me.ControllerContext.RequestContext, RouteTable.Routes, Function(c) c.List(listId, Nothing, Nothing))

            Return String.Concat(Me.Request.Url.Scheme, "://", Me.Request.ServerVariables("HTTP_HOST"), path)
        End Function

        Private Shared Function IsValidURL(ByVal url As String) As Boolean
            If url.ToLower().StartsWith("http://") OrElse url.StartsWith("https://") Then
                Return True
            End If

            Return False
        End Function

        Private Shared Function ShortenURL(ByVal url As String) As String
            If (Not IsValidURL(url)) Then
                Return String.Empty
            End If

            Dim requestUrl As String = String.Format("http://bit.ly/api?url={0}", url)

            Dim request As WebRequest = HttpWebRequest.Create(requestUrl)
            request.Proxy = Nothing
            Dim resultUrl As String = Nothing

            Try
                Using responseStream As Stream = request.GetResponse().GetResponseStream()
                    Dim reader As New StreamReader(responseStream, Encoding.ASCII)
                    resultUrl = reader.ReadToEnd()

                    If (Not IsValidURL(resultUrl)) Then
                        Dim w As New WebException(resultUrl)

                        Throw w
                    End If
                End Using
            Catch e1 As Exception
                Return url
            End Try

            If resultUrl.Length > url.Length Then
                resultUrl = url
            End If

            Return resultUrl
        End Function
    End Class
End Namespace