﻿// ----------------------------------------------------------------------------------
// Microsoft Developer & Platform Evangelism
// 
// Copyright (c) Microsoft Corporation. All rights reserved.
// 
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
// EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
// OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
// ----------------------------------------------------------------------------------
// The example companies, organizations, products, domain names,
// e-mail addresses, logos, people, places, and events depicted
// herein are fictitious.  No association with any real company,
// organization, product, domain name, email address, logo, person,
// places, or events is intended or should be inferred.
// ----------------------------------------------------------------------------------

namespace MyTodo.WebUx.Models
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Xml.Linq;
    using Microsoft.WindowsAzure;
    using Microsoft.WindowsAzure.ServiceRuntime;
    using Microsoft.WindowsAzure.StorageClient;
    using MyTodo.Data.WindowsAzure;

    public class TaskRepository
    {
        public virtual bool TablesExist()
        {
            var context = new TaskDataContext();
            return context.IsTablesCreated;
        }

        public virtual void CreateTables()
        {
            var context = new TaskDataContext();
            context.CreateTables();
        }

        public virtual TaskList CreateList(string name, bool isPublic)
        {
            var context = new TaskDataContext();
            var list = new TaskList { Name = name, IsPublic = isPublic };
            context.AddObject(TaskDataContext.TaskListTable, list.ToRow());
            context.SaveChanges();
           
            return list;
        }

        public virtual void CreateTask(Task task)
        {
            var context = new TaskDataContext();

            context.AddObject(TaskDataContext.TaskTable, task.ToRow());
            context.SaveChanges();
        }

        public virtual void DeleteTask(string listId, string taskId)
        {
            var context = new TaskDataContext();

            var task = context.Tasks
                .Where(t => t.PartitionKey == listId && t.RowKey == taskId && true)
                .SingleOrDefault();

            if (task != null)
            {
                context.DeleteObject(task);
                context.SaveChanges();
            }
        }

        public virtual void UpdateTask(Task task)
        {
            var context = new TaskDataContext();
            var taskRow = task.ToRow();

            context.AttachTo(TaskDataContext.TaskTable, taskRow, "*");
            context.UpdateObject(taskRow);
            context.SaveChanges();
        }

        public virtual Task GetTask(string taskId, string listId)
        {
            var context = new TaskDataContext();

            var task = context.Tasks
                .Where(t => t.RowKey == taskId && t.PartitionKey == listId && true)
                .ToModel()
                .SingleOrDefault();

            return task;
        }

        public virtual void UpdateList(TaskList list)
        {
            var context = new TaskDataContext();
            var taskListRow = list.ToRow();

            context.AttachTo(TaskDataContext.TaskListTable, taskListRow, "*");
            context.UpdateObject(taskListRow);
            context.SaveChanges();
        }

        public virtual IEnumerable<TaskList> GetTaskLists(bool includePrivate)
        {
            var context = new TaskDataContext();

            var lists = context.TaskLists
                .Where(t => t.IsPublic == true || t.IsPublic != includePrivate);

            return lists.ToModel();
        }

        public virtual TaskList GetTaskList(string listId)
        {
            var context = new TaskDataContext();

            if (this.TablesExist())
            {
                var list = context.TaskLists
                .Where(t => t.RowKey == listId && t.PartitionKey == listId.Substring(0, 1) && true)
                .ToModel()
                .SingleOrDefault();

                return list;
            }

            return null;
        }

        public virtual void DeleteList(string listId)
        {
            var context = new TaskDataContext();

            var list = context.TaskLists
               .Where(t => t.RowKey == listId && t.PartitionKey == listId.Substring(0, 1) && true)
               .SingleOrDefault();

            if (list != null)
            {
                var tasks = context.Tasks
                    .Where(t => t.PartitionKey == listId);

                foreach (TaskRow task in tasks)
                {
                    context.DeleteObject(task);
                }

                context.DeleteObject(list);
                context.SaveChanges();
            }
        }

        public PagedTaskSource GetTasks(string listId, int pageSize)
        {
            return this.GetTasks(listId, pageSize, null, null);
        }

        public virtual PagedTaskSource GetTasks(string listId, int pageSize, string nextPartition, string nextRowKey)
        {
            var context = new TaskDataContext();
            return new PagedTaskSource(context.Tasks.Where(t => t.ListId == listId), nextPartition, nextRowKey, pageSize);
        }

        public virtual XDocument RetrieveInitialLists()
        {
            var client = new WebClient();
            var blobUrl = RoleEnvironment.GetConfigurationSettingValue("InitialListsBlobUrl");

            try
            {
                using (var data = client.OpenRead(blobUrl))
                {
                    using (var reader = new StreamReader(data))
                    {
                        var blobContent = reader.ReadToEnd();

                        return XDocument.Parse(blobContent);
                    }
                }
            }
            catch
            {
            }

            return new XDocument();
        }

        public void CreatePublicBlobForInitialData(string containerName, string blobName)
        {
            CloudStorageAccount blobAccount = CloudStorageAccount.FromConfigurationSetting("DataConnectionString");

            CloudBlobClient blobStorage = blobAccount.CreateCloudBlobClient();
            blobStorage.RetryPolicy = RetryPolicies.Retry(1, TimeSpan.FromMilliseconds(100));

            CloudBlobContainer container = blobStorage.GetContainerReference(containerName.ToLowerInvariant());
            container.CreateIfNotExist();
            var permissions = container.GetPermissions();
            permissions.PublicAccess = BlobContainerPublicAccessType.Container;
            container.SetPermissions(permissions);

            CloudBlockBlob blob = container.GetBlockBlobReference(blobName);
            blob.UploadByteArray(new byte[0]);
        }
    }
}