﻿' ----------------------------------------------------------------------------------
' Microsoft Developer & Platform Evangelism
' 
' Copyright (c) Microsoft Corporation. All rights reserved.
' 
' THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
' EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
' OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
' ----------------------------------------------------------------------------------
' The example companies, organizations, products, domain names,
' e-mail addresses, logos, people, places, and events depicted
' herein are fictitious.  No association with any real company,
' organization, product, domain name, email address, logo, person,
' places, or events is intended or should be inferred.
' ----------------------------------------------------------------------------------

Imports System.Linq
Imports System.Reflection
Imports Microsoft.WindowsAzure
Imports Microsoft.WindowsAzure.StorageClient

Namespace Microsoft.WindowsAzure.StorageClient

    ''' <summary>
    ''' Development Storage currently requires the schema for an entity stored in a table to have been 
    ''' previously defined before you are allowed to query it. To provide the schema, the methods in this 
    ''' class create a new table, store a new entity with the required schema and then delete it. 
    ''' This workaround is applied only for Development Storage, Azure Storage does not have this restriction.
    '''
    ''' NOTE: These methods assume that entities derive from TableServiceEntity.
    ''' </summary>
    Public Class TableStorageExtensionMethods
        ''' <summary>
        ''' Creates a new table with the provided schema.
        ''' </summary>
        ''' <typeparam name="T">A TableServiceEntity derived class with the required schema.</typeparam>
        ''' <param name="tableStorage">The table storage client instance.</param>
        ''' <param name="entityName">The name of the table.</param>
        ''' <returns>true if the table was created, false if the table already existed.</returns>
        ''' <remarks>Use this method instead of the non-generic variant in StorageClient.</remarks>    
        Public Shared Function CreateTableIfNotExist(Of T As {TableServiceEntity, New})(ByVal tableStorage As CloudTableClient, ByVal entityName As String) As Boolean
            Dim result As Boolean = tableStorage.CreateTableIfNotExist(entityName)

            ' Execute conditionally for development storage only
            If tableStorage.BaseUri.IsLoopback Then
                InitializeTableSchemaFromEntity(tableStorage, entityName, New T())
            End If

            Return result
        End Function

        ''' <summary>
        ''' Creates tables from a DataServiceContext derived class. 
        ''' </summary>
        ''' <param name="serviceContextType">A DataServiceContext derived class that defines the table schemas.</param>
        ''' <param name="baseAddress">The baseAddress of the table storage endpoint.</param>
        ''' <param name="credentials">Storage account credentials.</param>
        ''' <remarks>
        ''' For each table, the class exposes one or more IQueryable&lt;T&gt; properties, where T is a 
        ''' TableServiceEntity derived class with the required schema.
        ''' </remarks>
        Public Shared Sub CreateTablesFromModel(ByVal serviceContextType As Type, ByVal baseAddress As String, ByVal credentials As StorageCredentials)
            CloudTableClient.CreateTablesFromModel(serviceContextType, baseAddress, credentials)

            Dim tableStorage As New CloudTableClient(baseAddress, credentials)

            ' Execute conditionally for development storage only
            If tableStorage.BaseUri.IsLoopback Then
                Dim ctx As TableServiceContext = tableStorage.GetDataServiceContext()

                Dim properties = serviceContextType.GetProperties(BindingFlags.Instance Or BindingFlags.Public)
                For Each table In properties.Where(Function(p) p.PropertyType.IsGenericType AndAlso p.PropertyType.GetGenericTypeDefinition() Is GetType(IQueryable(Of )))
                    Dim entity As TableServiceEntity = TryCast(Activator.CreateInstance(table.PropertyType.GetGenericArguments()(0)), TableServiceEntity)
                    If entity IsNot Nothing Then
                        InitializeTableSchemaFromEntity(tableStorage, table.Name, entity)
                    End If
                Next table
            End If
        End Sub

        Private Shared Sub InitializeTableSchemaFromEntity(ByVal tableStorage As CloudTableClient, ByVal entityName As String, ByVal entity As TableServiceEntity)
            Dim context As TableServiceContext = tableStorage.GetDataServiceContext()
            Dim now As DateTime = DateTime.UtcNow
            entity.PartitionKey = Guid.NewGuid().ToString()
            entity.RowKey = Guid.NewGuid().ToString()
            Array.ForEach(entity.GetType().GetProperties(BindingFlags.Public Or BindingFlags.Instance), Function(p) AnonymousMethod1(p, entity, now))

            context.AddObject(entityName, entity)
            context.SaveChangesWithRetries()
            context.DeleteObject(entity)
            context.SaveChangesWithRetries()
        End Sub

        Private Shared Function AnonymousMethod1(ByVal p As Object, ByVal entity As TableServiceEntity, ByVal now As DateTime) As Object
            If (p.Name <> "PartitionKey") AndAlso (p.Name <> "RowKey") AndAlso (p.Name <> "Timestamp") Then
                If p.PropertyType Is GetType(String) Then
                    p.SetValue(entity, Guid.NewGuid().ToString(), Nothing)
                ElseIf p.PropertyType Is GetType(DateTime) Then
                    p.SetValue(entity, now, Nothing)
                End If
            End If
            Return Nothing
        End Function
    End Class
End Namespace